/*
 * Pig
 * ICFP 2005
 * Drunk Sed Team
 * http://sed.free.fr/icfp2005
 */

#include "strategy.h"

#include "vars.h"

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

char *robber_name;
int world;
variables vars;
variables cops_names;
int nb_nodes;
node_type edges[MAX_NODES][MAX_NODES];
node_type hq;
node_type bank[NB_BANKS];
int money[NB_BANKS];
int my_money;
int cop[NB_COPS];

typedef struct {
  int node;
  int dist;
} node_dist;

typedef struct {
  node_dist *list;
  int maxsize;
  int head;
  int size;
} list;

list target_list;

void init_list(list *l)
{
  l->list = 0;
  l->maxsize = 0;
  l->size = 0;
  l->head = 0;
}

void free_list(list *l)
{
  if (l->maxsize)
    free(l->list);
  l->list = 0;
  l->maxsize = 0;
}

node_dist pop_head(list *);
void put_end(list *l, node_dist n)
{
  int pos;
  if (l->size == l->maxsize) {
    node_dist d;
    list l2;
    init_list(&l2);
    l2.list = malloc((l->maxsize + 1024) * sizeof(node_dist));
    if (!l2.list) {
      fprintf(stderr, "out of memory\n");
      exit(1);
    }
    l2.maxsize = l->maxsize + 1024;
    while (l->size) {
      d = pop_head(l);
      put_end(&l2, d);
    }
    free(l->list);
    memcpy(l, &l2, sizeof(list));
  }
  pos = l->head + l->size;
  if (pos >= l->maxsize) pos -= l->maxsize;
  l->list[pos] = n;
  l->size++;
}

node_dist pop_head(list *l)
{
  node_dist r;
  if (l->size == 0) {
    fprintf(stderr, "poping free list\n");
    exit(1);
  }
  r = l->list[l->head];
  l->head++;
  if (l->head == l->maxsize) l->head = 0;
  l->size--;
  return r;
}

node_dist read_tail(list *l)
{
  int pos;
  node_dist r;
again:
  if (l->size == 0) {
    r.dist = DELETED;
    return r;
  }
  pos = l->head + l->size - 1;
  if (pos >= l->maxsize) pos -= l->maxsize;
  r = l->list[pos];
  if (r.dist == DELETED) {
    l->size--;
    goto again;
  }
  return r;
}

node_dist pop_tail(list *l)
{
  node_dist r = read_tail(l);
  if (l->size) l->size--;
  return r;
}

int foot_distance[MAX_NODES];
int car_distance_from_hq[MAX_NODES];
int car_distance_to_hq[MAX_NODES];
int direct_distance[MAX_NODES];

/* type=FOOT => foot distance from hq
 * type=CAR & to_hq=0 => car distance from hq
 * type=CAR & to_hq=1 => car distance to hq
 */
void dist(int dist[MAX_NODES], int type, int to_hq)
{
  node_dist d;
  list l;
  int i;
  int cur;

  init_list(&l);

  d.node = hq;
  d.dist = 0;
  put_end(&l, d);

  while(l.size) {
    node_dist nd;
    d = pop_head(&l);

    /* if another path of length < no need to go by this way */
    if (dist[d.node] < d.dist)
      continue;

    dist[d.node] = d.dist;

    cur = d.node;

    if (to_hq == 1)
    for (i=0; i<nb_nodes; i++) {
      if (!(edges[i][cur] & type)) continue;

      nd.node = i;
      nd.dist = d.dist + 1;
      put_end(&l, nd);
    }
    else
    for (i=0; i<nb_nodes; i++) {
      if (!(edges[cur][i] & type)) continue;

      nd.node = i;
      nd.dist = d.dist + 1;
      put_end(&l, nd);
    }
  }

  free_list(&l);
}

/* get distance of every place if we keep on with our locomotion mode */
/* distance relative to cop number cp (0 is us) */
void direct_dist(int cp, int dist[MAX_NODES])
{
  node_dist d, nd;
  list l;
  int i;
  int type = cop[cp] > 0 ? FOOT : CAR;

  init_list(&l);

  d.node = abs(cop[cp]);
  d.dist = 0;
  put_end(&l, d);

  while(l.size) {
    d = pop_head(&l);

    /* if a shorter path, don't take this one */
    if (dist[d.node] < d.dist)
      continue;

    dist[d.node] = d.dist;

    for (i=0; i<nb_nodes; i++) {
      if (!(edges[d.node][i] & type)) continue;
      nd.node = i;
      nd.dist = d.dist + 1;
      put_end(&l, nd);
    }
  }

  free_list(&l);
}

/* return the next case the cop 'me' has to go
 * modifies the sign of cop[me] to reflect the way it has
 * to use
 */
int go_to(int me, int robber)
{
  list l;
  node_dist d, nd;
  int i;
  int dir_dist;
  int cut_dist;
  int cur_to_hq;
  int hq_to_robber;
  int by_foot;
  int length;
  int type;
  int *cop_dist;

  /* if already on it, return not move (same position, same locomotion) */
  if (robber == abs(cop[me]))
    return robber;

  init_list(&l);

  for (i=0; i<nb_nodes; i++)
    /* we put a value that is absolutely greater than the longest path */
    direct_distance[i] = MAX_NODES * 2;
  /* get the real dist from our position */
  direct_dist(me, direct_distance);
  /* so, how to go to this robber */
  /* direct way from our position? */
  dir_dist = direct_distance[robber];
  /* or is it faster if we go back to hq? */
  if (cop[me] > 0)
    cur_to_hq = foot_distance[cop[me]];
  else
    cur_to_hq = car_distance_to_hq[-cop[me]];
  if (car_distance_from_hq[robber] < foot_distance[robber]) {
    hq_to_robber = car_distance_from_hq[robber];
    by_foot = 0;
  } else {
    hq_to_robber = foot_distance[robber];
    by_foot = 1;
  }
  cut_dist = cur_to_hq + hq_to_robber;
  /* the big choice */
  if (dir_dist < cut_dist) {
direct:
    /* direct way */
    /* find the path */
    type = cop[me] > 0 ? FOOT : CAR;
    d.node = robber;
    put_end(&l, d);
    while (l.size) {
      d = pop_head(&l);
      length = direct_distance[d.node];
      for (i=0; i<nb_nodes; i++) {
        if (direct_distance[i] == length-1 && (edges[i][d.node] & type)) {
          if (i == abs(cop[me])) goto found;
          nd.node = i;
          put_end(&l, nd);
        }
      }
    }
    /* problem, we should not come here... */
    fprintf(stderr, "%s:%d:BUG!\n", __FILE__, __LINE__);
    /* anyway, do as if we chose not to move (maybe next turn
     * things will be ok)
     */
    free_list(&l);
    return abs(cop[me]);
found:
    free_list(&l);
    return d.node;
  }
  /* go back to hq */
  /* we might be on it */
  if (abs(cop[me]) == hq) {
    if (by_foot)
      cop[me] = abs(cop[me]);
    else
      cop[me] = -abs(cop[me]);
    /* might be useless calculation here */
    direct_dist(me, direct_distance);
    goto direct;
  }
  /* go to it */
  type = cop[me] > 0 ? FOOT : CAR;
  cop_dist = cop[me] > 0 ? foot_distance : car_distance_to_hq;
  d.node = hq;
  put_end(&l, d);
  while (l.size) {
    d = pop_head(&l);
    length = cop_dist[d.node];
    for (i=0; i<nb_nodes; i++) {
      if (cop_dist[i] == length+1 && (edges[i][d.node] & type)) {
        if (i == abs(cop[me])) goto found_hq;
        nd.node = i;
        put_end(&l, nd);
      }
    }
  }
  /* problem, we should not come here... */
  fprintf(stderr, "%s:%d:BUG!\n", __FILE__, __LINE__);
  /* anyway, do as if we chose not to move (maybe next turn
   * things will be ok)
   */
  free_list(&l);
  return abs(cop[me]);
found_hq:
  free_list(&l);
  return d.node;
}

void go_to_bank(void)
{
  int bk;
  int nb_loops = 1000000;
  /* a while(1) may fail, so let's do it this way */
  while (nb_loops) {
    nb_loops--;
    bk = rand() & 7;
    if (bk >= NB_BANKS) continue;
    if (abs(cop[0]) == bank[bk]) continue;
    if (abs(cop[1]) == bank[bk]) continue;
    if (abs(cop[2]) == bank[bk]) continue;
    if (abs(cop[3]) == bank[bk]) continue;
    if (abs(cop[4]) == bank[bk]) continue;
    break;
  }
  /* out of the loop without having chosen? probably never happens, but
   * who knows... let's dummy go to bank 0
   */
  if (bk >= NB_BANKS) bk = 0;
  add_target(bank[bk], BANK);
}

void move(void)
{
  int next;
  node_dist r;

  static int turn = 0;
  fprintf(stderr, "turn %d\n", turn++);

  /* strategy:
   * maintain a list of things to do, do the last one first
   * if the robber did bad things, go to the bank unless a cop
   * is already there
   * move around, dig dig dig
   * it's impossible to win alone, but I don't see how to cooperate
   * hey, I don't like cops, so let it behave like this, it's ok for me
   */
next_goal:
  r = read_tail(&target_list);
  if (r.dist == DELETED) {
    /* if no goal, go to bank */
    go_to_bank();
    //add_target(hq, BANK);
    r = read_tail(&target_list);
    //goto dont_move;
  }
  /* if goal is robber and a cop is there, nothing to win, skip this goal */
  if (r.dist == ROBBER_POSITION &&
      ( (abs(cop[1]) == r.node) ||
        (abs(cop[2]) == r.node) ||
        (abs(cop[3]) == r.node) ||
        (abs(cop[4]) == r.node))) {
    pop_tail(&target_list);
    goto next_goal;
  }
  /* if we achieved our goal, pop it */
  if (r.node == abs(cop[0])) {
    pop_tail(&target_list);
    goto next_goal;
  }
  /* if we are supposed to go to a bank where a cop is, don't go there */
  if (r.dist == BANK) {
    if (abs(cop[1]) == r.node ||
        abs(cop[2]) == r.node ||
        abs(cop[3]) == r.node ||
        abs(cop[4]) == r.node) {
      pop_tail(&target_list);
      goto next_goal;
    }
  }

  /* our goal seems reasonable, let's go */
  r = read_tail(&target_list);
  next = go_to(0, r.node);
  fprintf(stdout, "mov: %s cop-%s\n", get_variable_name(&vars, next),
          cop[0]>0?"foot":"car");
  fflush(stdout);
  return;

//dont_move:
  fprintf(stdout, "mov: %s cop-foot\n", get_variable_name(&vars, abs(cop[0])));
  fflush(stdout);
  return;
}

/* let's try to fool others, some people seem to code slow stuff
 * (according to the mailing list) and they may run out of time
 * trying to analyze this garbage
 */
void inform(void)
{
  int i;
  int nb = (int) (101.0*rand()/(RAND_MAX+1.0));
  /* i am a bit paranoid, you know */
  if (nb<0) nb = 0;
  if (nb>100) nb = 100;
  /* let's generate between 900 and 1000 random informations about robber */
  fprintf(stdout, "inf\\\n");
  if (robber_name)
  for (i=0; i<900+nb; i++) {
    int loc = (int) 1+((float)(nb_nodes-1)*rand()/(RAND_MAX+1.0));
    int world = (int) (200.0*rand()/(RAND_MAX+1.0));
    int certainty = (int)100- (201.0*rand()/(RAND_MAX+1.0));
    char *loc_name = get_variable_name(&vars, loc);
    /* still some paranoia */
    if (loc < 1) loc = 1;
    if (loc >= nb_nodes) loc = nb_nodes-1;
    if (world < 0) world = 0;
    if (world > 200) world = 200;
    if (certainty < -100) certainty = -100;
    if (certainty > 100) certainty = 100;
    if (!loc_name) loc_name = "55-and-woodlawn";
    fprintf(stdout, "inf: %s %s robber %d %d\n",
      robber_name, loc_name, world, certainty);
  }
  fprintf(stdout, "inf/\n");
  fflush(stdout);
}

void plan(void)
{
  node_dist r;
  /* default is to tell to cop i to go to bank i-1 (cop 0 is us) */
  /* but if a bank is corrupted, everyone goes to this bank !
   * (so that the robber can go to the other banks)
   */
  int mode = 0;
  int i;
  r = read_tail(&target_list);
  if (r.dist == ROBBER_POSITION) mode = 1;
  fprintf(stdout, "plan\\\n");
  for (i=1; i<NB_COPS; i++) {
    int next;
    if (mode == 0) /* default */
      next = go_to(i, bank[i-1]);
    else /* goto robber */
      next = go_to(i, r.node);
    fprintf(stdout, "plan: %s %s cop-%s %d\n",
      get_variable_name(&cops_names, i),
      get_variable_name(&vars, next),
      cop[i] > 0 ? "foot" : "car",
      world+1);
  }
  fprintf(stdout, "plan/\n");
  fflush(stdout);
}

/* democracy is best modeled as a random variable, so let's do it in
 * a democratic way (of course I vote for myself first)
 * (I guess I miss something here, cops and democracy, you know...)
 */
void vote(void)
{
  int i;
  int voted[5];
  int nb_loops;
  voted[1] = voted[2] = voted[3] = voted[4] = 0;
  fprintf(stdout, "vote\\\n");
  fprintf(stdout, "vote: %s\n", get_variable_name(&cops_names, 0));
  for (i=1; i < NB_COPS; i++) {
    /* normally it's a while(1), but I don't want to lose because of a
     * bad rand (let's lose by our own means)
     */
    nb_loops = 1000000;
    while(nb_loops) {
      nb_loops--;
      /* not the best random, but hey, it's democracy */
      /* we suppose that we have 4 other cops here */
      int x = rand() & 3;
      x++;
      if (voted[x]) continue;
      voted[x] = 1;
      fprintf(stdout, "vote: %s\n", get_variable_name(&cops_names, x));
      break;
    }
    if (nb_loops == 0)
      goto static_vote;
  }
end_vote:
  fprintf(stdout, "vote/\n");
  fflush(stdout);
  return;

static_vote:
  /* the nb_loops should be big enough, we should not come here */
  for(i=1; i<NB_COPS; i++) {
    if (voted[i]) continue;
    fprintf(stdout, "vote: %s\n", get_variable_name(&cops_names, i));
  }
  /* see the optimized usage of gotos! (Dijkstra is dead, you know)
   * (I never read his paper by the way)
   * (as far as I remember)
   * (yes this code is coming late in the contest, as you might expect)
   */
  goto end_vote;
}

void init_distances(void)
{
  int i;

  for (i=0; i<nb_nodes; i++) {
    /* we put a value that is absolutely greater than the longest path */
    foot_distance[i] = MAX_NODES * 2;
    car_distance_from_hq[i] = MAX_NODES * 2;
    car_distance_to_hq[i] = MAX_NODES * 2;
  }
  dist(foot_distance, FOOT, 0);
  dist(car_distance_from_hq, CAR, 0);
  dist(car_distance_to_hq, CAR, 1);
}

void add_target(int node, int type)
{
  node_dist r;
  int i;
  r.node = node;
  r.dist = type;
  if (type == ROBBER_POSITION) {
    /* delete previous ROBBER_POSITION (if there are some) */
    for (i=0; i<target_list.size; i++) {
      if (target_list.list[i].dist == ROBBER_POSITION) {
        target_list.list[i].dist = DELETED;
      }
    }
    /* add new one */
    put_end(&target_list, r);
    return;
  }
  /* everything else goes behind ROBBER_POSITION (if there is one) */
  put_end(&target_list, r);
  if (target_list.size == 1) return;
  /* if previous is ROBBER_POSITION, it must go head */
  if (target_list.list[target_list.size-2].dist == ROBBER_POSITION) {
    r = target_list.list[target_list.size-2];
    target_list.list[target_list.size-2] =
      target_list.list[target_list.size-1];
    target_list.list[target_list.size-1] = r;
  }
}

/* we might have smelled it, add all positions in range to target_list */
void smell(int dist)
{
  int i;
  int type = cop[0] > 0 ? FOOT : CAR;
  list d1;
  list d2;
  node_dist d;

  if (!dist) return;

  /* remove previous guesses, they are out of date */
  for (i=0; i < target_list.size; i++)
    if (target_list.list[i].dist == MY_GUESS)
      target_list.list[i].dist = DELETED;

  init_list(&d1);
  init_list(&d2);

  /* distance 1 */
  for (i=0; i<nb_nodes; i++) {
    if (!(edges[abs(cop[0])][i] & type)) continue;
    d.node = i;
    put_end(&d1, d);
  }

  /* distance 2 */
  if (dist == 2) {
    while (d1.size) {
      node_dist cur = pop_head(&d1);
      for (i=0; i<nb_nodes; i++) {
        if (!(edges[cur.node][i] & type)) continue;
        d.node = i;
        put_end(&d2, d);
      }
    }
  }

  /* put it in the todo list (with a go back to center, just in case */
  if (dist == 1) {
    while (d1.size) {
      d = pop_head(&d1);
      add_target(d.node, MY_GUESS);
      add_target(abs(cop[0]), MY_GUESS);
    }
  } else {
    while (d2.size) {
      d = pop_head(&d2);
      add_target(d.node, MY_GUESS);
      add_target(abs(cop[0]), MY_GUESS);
    }
  }

  free_list(&d1);
  free_list(&d2);
}

/* seek around node */
void around(int node)
{
  int i;
  list d1;
  node_dist d;

  init_list(&d1);

  /* we add the current node too to go back there (might be better
   * behavior if by foot, might be worst if by car, but hey, it's
   * just a stupid cop)
   */
  /* distance 1 */
  for (i=0; i<nb_nodes; i++) {
    if (!edges[node][i]) continue;
    d.node = i;
    add_target(i, AROUND_GUESS);
    add_target(node, AROUND_GUESS);
    put_end(&d1, d);
  }

  /* distance 2 */
  while (d1.size) {
    node_dist cur = pop_head(&d1);
    for (i=0; i<nb_nodes; i++) {
      if (!edges[cur.node][i]) continue;
      add_target(i, AROUND_GUESS);
      add_target(node, AROUND_GUESS);
    }
  }

  free_list(&d1);
}
