/*
 * A hash table - more or less generic
 *---
 * to my angel
 * public domain
 * march 2002
 */

#ifndef _HASH_H_
#define _HASH_H_

typedef unsigned long hid;
typedef void new_data;
typedef void hash_data;

/* flag values */
#define	HASH_NO_ERROR		0
#define HASH_OUT_OF_MEMORY	1
#define HASH_ALREADY_PRESENT	2
#define HASH_ALREADY_FREE	3
#define HASH_NOT_PRESENT	4

/* must be power of 2 */
#define HASH_INIT_SIZE_SMALL	4096
#define HASH_INIT_SIZE_BIG	(256 * 1024)

typedef int(*hash_compare_function)(void *, new_data *, hash_data *);
typedef unsigned long (*hash_hash_function)(void *, new_data *);
typedef new_data *(*hash_get_key_function)(void *, hash_data *);

typedef struct {
  /* data for the hash table */
  void *data;
  unsigned char *data_busy;
  int elem_size;
  int flag;
  unsigned long size;
  unsigned long maxsize;
  hash_compare_function compare;
  void *compare_pointer;
  hash_hash_function hash;
  void *hash_pointer;
  hash_get_key_function get_key;
  void *get_key_pointer;
  /* data for the hash function */
  unsigned long hash_current;
  unsigned long hash_random;
  unsigned long size_current;
  unsigned long hash_collision;
} hash_table;

#define hash_get_elem_size(h) ((h)->elem_size)

/* functions for hash table manipulation */
void init_hash(hash_table *, int, int,
	       hash_compare_function, void *,
	       hash_hash_function, void *,
	       hash_get_key_function, void *);
void free_hash(hash_table *);
hid hash_get_new_slot(hash_table *, new_data *);
hid hash_get_slot(hash_table *, new_data *);
void hash_feed_slot(hash_table *, hid, hash_data *);
void hash_free_slot(hash_table *, hid);
hash_data *hash_get_slot_data(hash_table *, hid);

/* functions for hash value calculation */
void hash_start(hash_table *);
void hash_update(hash_table *, void *, unsigned long);
unsigned long hash_end(hash_table *);
unsigned long hash_simple(hash_table *, void *, unsigned long);

/* default functions for compare, hash and get_key */
#define HASH_DEFAULT_COMPARE		(hash_compare_function)hash_default_compare
#define HASH_DEFAULT_HASH		(hash_hash_function)hash_default_hash
#define HASH_DEFAULT_GET_TRUE_DATA	hash_default_get_key
int hash_default_compare(hash_table *, new_data *, hash_data *);
unsigned long hash_default_hash(hash_table *, new_data *);
new_data *hash_default_get_key(void *, hash_data *);

#endif /* _HASH_H_ */
