/*
 * A mini transition relation postfix reader
 * ---
 * January 2002
 * to my angel
 */

#include "vars.h"

#include <string.h>
#include <stdio.h>

#include "mem.h"
#include "hash.h"
#include "strings.h"

#define SIZEINC	16

static sid user_to_sid(variables *v, variable *var)
{
#ifdef DEBUG
  if (*var < 0 || *var >= v->size) {
    fprintf(stderr, "vars.c:user_to_sid (line %d): error : *var = %d\n", __LINE__, *var);
    abort();
  }
#endif

  return v->data[*var].name;
}

/* return 0 if ok, -1 if pb */
int init_vars(variables *v)
{
  v->data = 0;
  v->size = v->maxsize = 0;

  return init_strings(&v->strings, (string_user_to_sid_function)user_to_sid, (user_context *)v, sizeof(variable));
}

void free_vars(variables *v)
{
  free_strings(&v->strings);

  if (v->maxsize)
    freemem(v->data);

  v->size = v->maxsize = 0;
  v->data = 0;
}

/* return NO_VARIABLE on error, the newly created var if all ok
 * will return NO_VARIABLE if the var already exist
 */
variable new_variable(variables *v, char *s)
{
  sid sd;
  hid hd;
  variable d;

  if(string_new(&v->strings, s, &hd, &sd)!=0)
    return NO_VARIABLE;

  d = v->size;
  string_feed_slot(&v->strings, hd, (void *)&d);

  if (v->size == v->maxsize)
    v->data = regetmem(v->data, sizeof(_variable) * (v->maxsize+=SIZEINC));

  memset(&v->data[v->size], 0, sizeof(_variable));

  v->data[v->size].type = VARIABLE_TYPE_NO_TYPE;
  v->data[v->size].name = sd;
  v->data[v->size].bdd1 = 0;
  v->data[v->size].bdd2 = 0;
  v->data[v->size].canonik_variable = NO_VARIABLE;
  v->data[v->size].trans_fun = 0;
  v->data[v->size].tag = 0;

  v->size++;

  return d;
}

/* return NO_VARIABLE on error, the newly created var if all ok
 * will return the var if it exists
 * sets *new to -1 if error, 0 if var already exists, 1 if it was created
 */
variable new_or_get_variable(variables *v, char *s, int *new)
{
  sid sd;
  hid hd;
  variable d;
  int ret;

  *new = -1;

  if((ret=string_new(&v->strings, s, &hd, &sd))==-1)
    return NO_VARIABLE;

  *new = 0;

  if (ret == -2)
    return *(variable *)string_get_slot(&v->strings, hd);

  *new = 1;

  d = v->size;
  string_feed_slot(&v->strings, hd, (void *)&d);

  if (v->size == v->maxsize)
    v->data = regetmem(v->data, sizeof(_variable) * (v->maxsize+=SIZEINC));

  memset(&v->data[v->size], 0, sizeof(_variable));

  v->data[v->size].type = VARIABLE_TYPE_NO_TYPE;
  v->data[v->size].name = sd;
  v->data[v->size].bdd1 = 0;
  v->data[v->size].bdd2 = 0;
  v->data[v->size].canonik_variable = NO_VARIABLE;
  v->data[v->size].trans_fun = 0;
  v->data[v->size].tag = 0;

  v->size++;

  return d;
}
/* return the var if it exists, NO_VARIABLE if not */
variable get_variable(variables *v, char *s)
{
  sid sid;
  variable *var;

  /* gcc3.3 doesn't like "(user_data **)&var"                           */
  /* if (string_search(&v->strings, s, &sid, (user_data **)&var) == -1) */
  if (string_search(&v->strings, s, &sid, (void *)&var) == -1)
    return NO_VARIABLE;

  return *var;
}

char *get_variable_name(variables *v, variable x)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return 0;
  }

  return sid_to_string(&v->strings, v->data[x].name);
}

void *get_variable_bddvar(variables *v, variable x)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return 0;
  }

  return v->data[x].bdd1;
}

void set_variable_bddvar(variables *v, variable x, void *y)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return;
  }

  v->data[x].bdd1 = y;
}

void *get_variable_bdd(variables *v, variable x)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return 0;
  }

  return v->data[x].bdd2;
}

void set_variable_bdd(variables *v, variable x, void *y)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return;
  }

  v->data[x].bdd2 = y;
}

void set_variable_type(variables *v, variable x, int type)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return;
  }

  v->data[x].type = type;
}

void add_variable_type(variables *v, variable x, int type)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return;
  }

  v->data[x].type |= type;
}

int get_variable_type(variables *v, variable x)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return VARIABLE_TYPE_NO_TYPE;
  }

  return v->data[x].type;
}

void set_variable_canon(variables *v, variable x, variable canon )
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return;
  }

  v->data[x].canonik_variable = canon;
}

variable get_variable_canon(variables *v, variable x)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return NO_VARIABLE;
  }

  return v->data[x].canonik_variable;
}

void *get_variable_trans_fun(variables *v, variable x)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return 0;
  }

  return v->data[x].trans_fun;
}

void set_variable_trans_fun(variables *v, variable x, void *d)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return;
  }

  v->data[x].trans_fun = d;
}

int get_variable_tag(variables *v, variable x)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return 0;
  }

  return v->data[x].tag;
}

void set_variable_tag(variables *v, variable x, int d)
{
  if (x<0 || x>= v->size) {
    fprintf(stderr, "%s:%d:%s: out of bounds access of variables array\n", __FILE__, __LINE__, __FUNCTION__);
    return;
  }

  v->data[x].tag = d;
}
