/*
 * Robber
 * ICFP 2005
 * Drunk Sed Team
 * http://sed.free.fr/icfp2005
 */

#include "verbose.h"
#include "vars.h"
#include "strategy.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

char *line;
int line_size;
int line_maxsize;

int get_char(void)
{
  int c = getchar();
  if (c == EOF) {
    fprintf(stderr, "error, EOF\n");
    exit(1);
  }
  return c;
}

void put_char(int c)
{
  if (line_size == line_maxsize) {
    line = realloc(line, line_maxsize += 128);
    if (!line) {
      fprintf(stderr, "out of memory\n");
      exit(1);
    }
  }

  line[line_size] = c;
  line_size++;
}

void get_token(void)
{
  int c;

  line_size = 0;

  do c = get_char(); while (isspace(c));

  while (!isspace(c)) {
    put_char(c);
    c = get_char();
  }

  put_char(0);
}

void wsk(void)
{
  int bk = 0;
  int nb_edges = 0;

  /* name */
  get_token(); get_token();
  fprintf(stderr, "my name: %s\n", line);

  /* robber */
  get_token(); get_token();
  fprintf(stderr, "the robber: %s\n", line);

  /* 5 cops */
  get_token(); get_token();
  get_token(); get_token();
  get_token(); get_token();
  get_token(); get_token();
  get_token(); get_token();

  /* nodes */
  get_token();
  if (strcmp(line, "nod\\")) {
    fprintf(stderr, "nod\\ expected\n");
    exit(0);
  }
  while (1) {
    int node;
    get_token();
    /* end of nodes ? */
    if (!strcmp(line, "nod/"))
      break;
    /* node name */
    get_token();
    node = new_variable(&vars, line);
    if (node == -1) {
      fprintf(stderr, "error creating node\n");
      exit(1);
    }
    if (node >= MAX_NODES) {
      fprintf(stderr, "too much nodes for me, I guess I am disqualified...\n");
      exit(1);
    }
    nb_nodes++;
    /* node tag */
    get_token();
    if (!strcmp(line, "hq"))
      hq = node;
    if (!strcmp(line, "bank")) {
      if (bk == NB_BANKS) {
        fprintf(stderr, "too much banks\n");
        exit(1);
      }
      bank[bk] = node;
      bk++;
    }
    if (!strcmp(line, "robber-start"))
      robber = node;
    /* coordinates */
    get_token(); get_token();
  }

  /* edges */
  get_token();
  if (strcmp(line, "edg\\")) {
    fprintf(stderr, "edg\\ expected\n");
    exit(0);
  }
  while (1) {
    int node1;
    int node2;
    get_token();
    /* end of edges ? */
    if (!strcmp(line, "edg/"))
      break;
    /* start node */
    get_token();
    node1 = get_variable(&vars, line);
    if (node1 == -1) {
      fprintf(stderr, "%s: node not found\n", line);
      exit(1);
    }
    /* end node */
    get_token();
    node2 = get_variable(&vars, line);
    if (node2 == -1) {
      fprintf(stderr, "%s: node not found\n", line);
      exit(1);
    }
    /* edge type */
    get_token();
    if (!strcmp(line, "car")) {
      edges[node1][node2] |= CAR;
    } else if (!strcmp(line, "foot")) {
      edges[node1][node2] |= FOOT;
      edges[node1][node2] |= CAR;
      edges[node2][node1] |= FOOT;
    } else {
      fprintf(stderr, "%s: bad edge type\n", line);
      exit(1);
    }
    nb_edges++;
  }
  /* wsk/ */
  get_token();
  if (strcmp(line, "wsk/")) {
    fprintf(stderr, "%s: expected wsk/\n", line);
    exit(1);
  }
}

void wor(void)
{
  int i;

  /* world */
  get_token(); get_token();
  /* rbd */
  get_token(); get_token();
  my_money = atoi(line);
  /* bv */
  get_token();
  for (i=0; i<NB_BANKS; i++) {
    int bk;
    int j;
    get_token(); get_token();
    bk = get_variable(&vars, line);
    if (bk == -1) {
      fprintf(stderr, "no bank at %s\n", line);
      exit(1);
    }
    for (j=0; j<NB_BANKS; j++)
      if (bank[j] == bk)
        break;
    if (j == NB_BANKS) {
      fprintf(stderr, "%s:%d:should not read this man\n", __FILE__, __LINE__);
      exit(1);
    }
    get_token();
    money[j] = atoi(line);
  }
  get_token();
  /* ev */
  get_token(); get_token();
  /* smell */
  get_token(); get_token();
  /* pl */
  get_token();
  if (strcmp(line, "pl\\")) {
    fprintf(stderr, "%s: expected pl\\\n", line);
    exit(1);
  }
  i = 0;
  while (1) {
    int node;
    get_token();
    if (!strcmp(line, "pl/"))
      break;
    /* bot */
    get_token();
    /* location */
    get_token();
    node = get_variable(&vars, line);
    if (node == -1) {
      fprintf(stderr, "%s: unknown location\n", line);
      exit(1);
    }
    /* ptype */
    get_token();
    if (!strcmp(line, "robber")) {
      robber = node;
    } else if (!strcmp(line, "cop-foot")) {
      if (i == NB_COPS) {
        fprintf(stderr, "too much cops\n");
        exit(1);
      }
      cop[i] = node;
      i++;
    } else if (!strcmp(line, "cop-car")) {
      if (i == NB_COPS) {
        fprintf(stderr, "too much cops\n");
        exit(1);
      }
      cop[i] = -node;
      i++;
    } else {
      fprintf(stderr, "%s: unknown bot type\n", line);
      exit(1);
    }
  }
  /* wor/ */
  get_token();
  if (strcmp(line, "wor/")) {
    fprintf(stderr, "%s: wor/ expected\n", line);
    exit(1);
  }
}

void get_message(void)
{
  get_token();

  if (!strcmp(line, "wsk\\")) {
    wsk();
    //init_distances();
    return;
  }

  if (!strcmp(line, "wor\\")) {
    wor();
    move();
    return;
  }

  if (!strcmp(line, "game-over")) {
    fprintf(stderr, "game over\n");
    exit(0);
  }
}

int main(int n, char **v)
{
  //int v;
  verbose verb;

  if (n == 1) fclose(stderr);

  /* for no gcc warning */
  if (0) { verbose *f = &verb; f++; }

  /* we require this */
  if (sizeof(void *) < sizeof(long)) {
    fprintf(stderr, "sizeof(void *) < sizeof(long) here, I can't run.\n");
    return 1;
  }

  if (init_verbose(&verb) == -1) {
    fprintf(stderr, "fatal error : can't init verbose mode\n");
    return 1;
  }

  if (init_vars(&vars) == -1) {
    fprintf(stderr, "error init vars\n");
    return 1;
  }

  /* node 0 must not exist */
  if (new_variable(&vars, ". node 0 @ |") == -1) {
    fprintf(stderr, "internal error\n");
    return 1;
  }
  nb_nodes++;

  fprintf(stdout, "reg: -)DrunkSedRobber#( robber\n");
  fflush(stdout);

  while (1) {
    get_message();
  }

  return 0;
}
