/*
 * Robber
 * ICFP 2005
 * Drunk Sed Team
 * http://sed.free.fr/icfp2005
 */

#include "strategy.h"

#include "vars.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

variables vars;
int nb_nodes;
node_type edges[MAX_NODES][MAX_NODES];
node_type hq;
node_type bank[NB_BANKS];
node_type robber;
int money[NB_BANKS];
int my_money;
int cop[NB_COPS];

typedef struct {
  int node;
  int dist;
} node_dist;

typedef struct {
  node_dist *list;
  int size;
  int maxsize;
  int head;
} list;

void init_list(list *l)
{
  l->list = 0;
  l->maxsize = 0;
  l->size = 0;
  l->head = 0;
}

void free_list(list *l)
{
  if (l->maxsize)
    free(l->list);
  l->list = 0;
  l->maxsize = 0;
}

node_dist pop_head(list *);
void put_end(list *l, node_dist n)
{
  int pos;
  if (l->size == l->maxsize) {
    node_dist d;
    list l2;
    init_list(&l2);
    l2.list = malloc((l->maxsize + 1024) * sizeof(node_dist));
    if (!l2.list) {
      fprintf(stderr, "out of memory\n");
      exit(1);
    }
    l2.maxsize = l->maxsize + 1024;
    while (l->size) {
      d = pop_head(l);
      put_end(&l2, d);
    }
    free(l->list);
    memcpy(l, &l2, sizeof(list));
  }
  pos = l->head + l->size;
  if (pos >= l->maxsize) pos -= l->maxsize;
  l->list[pos] = n;
  l->size++;
}

node_dist pop_head(list *l)
{
  node_dist r;
  if (l->size == 0) {
    fprintf(stderr, "poping free list\n");
    exit(1);
  }
  r = l->list[l->head];
  l->head++;
  if (l->head == l->maxsize) l->head = 0;
  l->size--;
  return r;
}

int cop_distance[MAX_NODES];
int bank_distance[MAX_NODES];
int robber_distance[MAX_NODES];

void cop_dist(int cop, int dist[MAX_NODES])
{
  node_dist d;
  list l;
  int i;
  int by_car = 0;

  init_list(&l);

  d.node = cop;
  d.dist = 0;
  put_end(&l, d);

  /* if on hq, the cop may go by foot or car */
  if (abs(cop) == hq) {
    d.node = -cop;
    d.dist = 0;
    put_end(&l, d);
  }

  while(l.size) {
    node_dist nd;
    d = pop_head(&l);

    /* if another path of length < no need to go by this way */
    if (dist[abs(d.node)] < d.dist)
      continue;

    dist[abs(d.node)] = d.dist;

    cop = d.node;

    by_car = 0;
    if (cop < 0) { cop = -cop; by_car++; }

    for (i=0; i<nb_nodes; i++) {
      if (!edges[cop][i]) continue;

      /* if the pig is on the hq, he may go by foot or car */
      if (i == hq) {
        if (edges[cop][i] & CAR) {
          nd.node = -i;
          nd.dist = d.dist + 1;
          put_end(&l, nd);
        }
        if (edges[cop][i] & FOOT) {
          nd.node = i;
          nd.dist = d.dist + 1;
          put_end(&l, nd);
        }
        continue;
      }

      /* is this fat pig moving with a car? */
      if (by_car && (edges[cop][i] & CAR)) {
        nd.node = -i;
        nd.dist = d.dist + 1;
        put_end(&l, nd);
        continue;
      }
      /* or is he on his feet (poor little pig)? */
      if (!by_car && (edges[cop][i] & FOOT)) {
        nd.node = i;
        nd.dist = d.dist + 1;
        put_end(&l, nd);
        continue;
      }
    }
  }

  free_list(&l);
}

void bank_dist(int bank, int dist[MAX_NODES])
{
  node_dist d, nd;
  list l;
  int i;

  init_list(&l);

  d.node = bank;
  d.dist = 0;
  put_end(&l, d);

  while(l.size) {
    d = pop_head(&l);

    /* if a shorter path, don't take this one */
    if (dist[d.node] < d.dist)
      continue;

    dist[d.node] = d.dist;

    for (i=0; i<nb_nodes; i++) {
      if (!(edges[i][d.node] & FOOT)) continue;
      nd.node = i;
      nd.dist = d.dist + 1;
      /* if a cop near, it's dangerous to go this way */
      if (cop_distance[i] == 2)
        nd.dist += 1;
      if (cop_distance[i] == 1)
        nd.dist += 2;
      if (cop_distance[i] == 0)
        nd.dist += 3;
      put_end(&l, nd);
    }
  }

  free_list(&l);
}

void move(void)
{
  int i;
  int next;
  int nb_targets;

  static int turn = 0;
  fprintf(stderr, "turn %d\n", turn++);

  for (i=0; i<nb_nodes; i++) {
    /* we put a value that is absolutely greater than the longest path */
    cop_distance[i] = MAX_NODES * 2;
    bank_distance[i] = MAX_NODES * 2;
  }

  for (i=0; i<NB_COPS; i++)
    cop_dist(cop[i], cop_distance);

  nb_targets = 0;
  for (i=0; i<NB_BANKS; i++) {
    /* this bank is meaningful only if it has or will have money
     * and no cop too near
     */
    /* don't take into account this silly bank in the cul-de-sac
     * if a pig is too near (let's say 3 steps) (of course a well
     * coordinated team could beat me here, but I don't think such
     * a team will ever exist unless some people submit several bots.
     * let's hardcode (yeah it's bad, but it's sunday, deadline is in 24h
     * tomorrow I don't code, so... let's remain straightforward)
     */
    int pos = get_variable(&vars, "58-culdesac");
    if (pos == -1)
      fprintf(stderr, "58-culdesac does not exist\n");
    if (bank[i] != pos || cop_distance[pos] > 3)
      if ((money[i] != 0 || my_money < 5975) && cop_distance[bank[i]] > 1) {
        bank_dist(bank[i], bank_distance);
        nb_targets++;
      }
  }
  if (!nb_targets) {
    /* let's keep moving, select a target far from cops */
    int i;
    int dist = 0;
    int node = -1;
    for (i=1; i<nb_nodes; i++) {
      /* farther is better */
      if (cop_distance[i] > dist) {
        dist = cop_distance[i];
        node = i;
      } else if (cop_distance[i] == dist) {
        /* randomly choose to keep this target or to go to the new one */
        if (rand() & 1) {
          dist = cop_distance[i];
          node = i;
        }
      }
    }
fprintf(stderr, "moving to %s (dist %d)\n", get_variable_name(&vars, node),
  dist);
    if (node != -1) /* should pass, but we never know... */
      bank_dist(node, bank_distance);
  }

  /* make a step to go to the nearest bank */
  /* default: no move */
  next = robber;
  /* go to a bank */
  for (i=0; i<nb_nodes; i++) {
    if (!(edges[robber][i] & FOOT))
      continue;
    /* would we get closer? */
    if (bank_distance[i] <= bank_distance[next]) {
      /* if a cop is there, don't go */
      if (cop_distance[i] == 0)
        continue;
      /* if it's a bank and there is a cop near, don't go */
      if (bank_distance[i] == 0 && cop_distance[i] <= 1)
        continue;
      /* if same distance, take the one farther from cop */
      if (bank_distance[i] == bank_distance[next] &&
          cop_distance[i] < cop_distance[next])
        continue;
      /* try it */
      next = i;
    }
  }
  /* if we are on a bank and don't move, we must watch cops around */
  if (next == robber && bank_distance[next] == 0 && cop_distance[next] <= 1) {
    /* cop near, time to move */
    for (i=0; i<nb_nodes; i++) {
      if (!(edges[robber][i] & FOOT))
        continue;
      /* if the cop distance is bigger, it's good,
       * we try to go near a bank in the process, $ is good for us
       * if the distance is the same than the robber pos, move, cops see us
       */
      if (cop_distance[i] > cop_distance[next] ||
         (cop_distance[i] == cop_distance[next] &&
          bank_distance[i] < bank_distance[next]) ||
         (cop_distance[i] == cop_distance[next] &&
          next == robber))
        next = i;
    }
  }
  /* ok, let's go with that, if we die, we die... */

fprintf(stderr, "mov to %s (cop distance %d)\n", get_variable_name(&vars, next),
cop_distance[next]);
  fprintf(stdout, "mov: %s robber\n", get_variable_name(&vars, next));
  fflush(stdout);
}
